import thorpy, math, pygame
from pygame.math import Vector2 as V
from parameters import MIN_D, GAME_RECT, W, H, MULT_EXPLO, MULT_FIRE, MIN_MASS
from parameters import R_SMALL, R_MED, R_BIG, COLL_CONST
import predefined


class CelestialBody(object):
    current_id = 0

    def __init__(self, r, m, q, color, v=None, img=None):
        self.r = r
        self.m = m
        self.q = V(q)
        if v is None: v=V()
        self.v = v
        self.color = color
        if not img:
            self.img = self.build_image()
        else:
            self.img = img
        self.burning = False
        self.explosive = False
        self.brother = None
        self.smoke = None
        self.debris = None
        self.vr = V(self.r,self.r)
        self.id = CelestialBody.current_id
        self.is_asteroid = False
        self.must_death = False
        CelestialBody.current_id += 1

    def build_image(self):
        size = (2*self.r, 2*self.r)
        base = thorpy.graphics.get_aa_round_rect(size, self.r, self.color)
        #todo : star shadow
        return base

    def copy(self): #doesnt copy id!!
        body = CelestialBody(self.r, self.m, self.q, self.color, V(self.v),
                                self.img)
        body.burning = self.burning
        body.explosive = self.explosive
        body.is_asteroid = self.is_asteroid
        body.must_death = self.must_death
        if self.smoke:
            body.smoke = self.smoke.get_copy()
        if self.debris:
            body.debris = self.debris.get_copy()
        return body


class Universe(object):

    def __init__(self, G, dt, bodies):
        self.G = G
        self.bodies = bodies
        self.dt = dt
        self.BASEDT = dt
        self.must_explode = set([])
        self.kill_smokes = []
        self.kill_debris = set([])

    def get_copy(self):
        """sets and lists are emptied!"""
        bodies = [b.copy() for b in self.bodies]
        return Universe(self.G, self.dt, bodies)

    def update_velocities(self):
        for b1 in self.bodies:
            for b2 in self.bodies:
                if b1 is not b2:
                    vect = b2.q - b1.q
                    d = vect.length()
                    if d < MIN_D:
                        d = MIN_D
                        direction = V(random.random(),random.random()).normalize()
                        F = direction * self.G * b1.m * b2.m / (d*d)
                    else:
                        F = vect.normalize() * self.G * b1.m * b2.m / (d*d)
                    a = F / b1.m
                    b1.v += a * self.dt

    def update_brothers(self):
        for b in self.brothers:
            b.q = b.brother + b.broq

    def update_positions(self):
        for b in self.bodies:
            b.q += b.v * self.dt

    def draw(self, ap):
        for b in self.bodies:
            ap.blit(b.img, b.q-b.vr)

    def orbit(self, body, parent, a=None, spec_energy=None, sign=1):
        """
        Set the velocity in order to obtain circular orbit, at current position.
        """
        assert sign == 1 or sign == -1
        r = parent.q - body.q
        d = r.length()
        #default : circular
        mu = parent.m*self.G
        assert not(a and spec_energy)
        if a is None: a = d #maxbe ignored (see below)
        if spec_energy is None: spec_energy = -mu/(2.*a)
        body.v = r.rotate(sign*90)
        vnorm = math.sqrt(2*(mu/d + spec_energy))
        body.v.scale_to_length(vnorm)
        body.v += parent.v

    def is_orbiting(self, body, parent):
        r = parent.q - body.q
        d = r.length()
        mu = parent.m*self.G
        spec_energy = 0.5*body.v.length()**2 - mu/d
        if spec_energy < 0:
            return True
        return False

    def center_on(self, body):
        delta = V(W//2,H//2) - body.q
        for b in self.bodies:
            b.q += delta

##    def kill(self, body, q, body2=None):
##        print("Killing", body.id)
##        try:
##            self.bodies.remove(body)
##        except:
##            pass
##        v = tuple(body.v)
##        self.must_explode.add((body,q,v))
##        self.kill_debris.add(body)
##        if body.smoke:
##            self.kill_smokes.append((body.smoke, tuple(body.q)))
##        if body2:
##            self.must_explode.add((body2,q,v))
##            self.kill_debris.add(body2)

    def kill(self, body, q, body2=None):
        body.must_death = True
        v = tuple(body.v)
        self.must_explode.add((body,q,v))
        self.kill_debris.add(body)
        if body.smoke:
            self.kill_smokes.append((body.smoke, tuple(body.q)))
        if body2:
            self.must_explode.add((body2,q,v))
            self.kill_debris.add(body2)

    def collide(self, b1, b2):
        """Here is the logic:
            * the lightest body always die ...
            * ... so we can directly compute the mass lost for heaviest one.
        """
        r = b2.q - b1.q
        d = r.length()
        if d < b1.r + b2.r:
            M = b1.m + b2.m
            dv = (b1.v - b2.v).length()
            b1s = V(r)
            b1s.scale_to_length(b1.r)
            contact = tuple(b1.q + b1s)
            if b1.m < b2.m:
                light = b1
                heavy = b2
            else:
                light = b2
                heavy = b1
            heavy.v = heavy.m*heavy.v/M + light.m*light.v/M
            self.kill(light, contact, heavy)
            if heavy.explosive: #if heavy is explosive, collision explode it
                self.kill(heavy, contact)
                return
            multiplier = 1.
##            if light.explosive: multiplier *= MULT_EXPLO
            if light.burning: multiplier *= MULT_FIRE
            if heavy.burning : multiplier /= MULT_FIRE
            heavy.m -= light.m * multiplier * COLL_CONST * dv
            if heavy.m < MIN_MASS:
                self.kill(heavy, contact)

    def collide_all(self):
        i = 0
        collisions = []
        L = len(self.bodies)
        while i+1 < L:
            b1 = self.bodies[i]
            for b2 in self.bodies[i+1:]:
                if (b1.q - b2.q).length() < (b1.r+b2.r):
                    self.collide(b1,b2)
##                    collisions.append((b1,b2))
            i += 1
##        for b1,b2 in collisions:
##            self.collide(b1,b2)